# Attempt at a Minesweeper game in TCL
# George Ruban
# July 10, 2000
# Dec 11, 2001 - implemented menus, bitmaps?

set fgcolor() black
set fgcolor(#) black
set fgcolor(1) blue
set fgcolor(2) green
set fgcolor(3) red
set fgcolor(4) purple
set fgcolor(5) violet
set fgcolor(6) $fgcolor(5)
set fgcolor(7) $fgcolor(5)
set fgcolor(8) $fgcolor(5)

set rows 8
set columns 8
set mines 10
set minesLeft $mines
set timer 0

# width for images is measured in pixels, for text in characters.
set wImage 20
set wText 2

# Counts adjacent mines to input coordinate, returns # on a hit, "" on 0,
# otherwise number of adjacent mines. Displayed when button clicked.
proc countMines {x y} {
    global mine
    if {$mine($x,$y)} {return #}
    set retval 0
    for {set i [expr $x-1]} {$i <= $x+1} {incr i} {
	    for {set j [expr $y-1]} {$j <= $y+1} {incr j} {
            if {[info exists mine($i,$j)] && $mine($i,$j)} {incr retval}
        }
    }
    if $retval {return $retval} {return ""}
}

# Called when a minefield is left-clicked to display what is underneath
proc reveal {x y} {
    global .b mine fgcolor gameOver wImage
    global columns rows mines revealed
    if {$gameOver} return
    set relief [lindex [.b($x,$y) configure -relief] 4]
    if {$relief != "raised"} return
    set text [lindex [.b($x,$y) configure -text] 4]
    if {$text == "F"} return
    set show [countMines $x $y]
    .b($x,$y) configure -text $show -relief ridge -foreground $fgcolor($show)
    if {$show=="#"} { 
        # Hit a mine, player loses, expose all mines
        .b($x,$y) configure -background red
        set gameOver 1
        for {set i 0} {$i < $columns} {incr i} {
    	    for {set j 0} {$j < $rows} {incr j} {
                if {$mine($i,$j)} {
                    .b($i,$j) configure -image mineImage -width $wImage -text "#" -relief ridge -foreground $fgcolor(#)
                }
            }
        }
        return
    } 
    incr revealed
    if {$revealed >= [expr $columns*$rows-$mines]} {
        # won, game over
        set gameOver 1
        for {set i 0} {$i < $columns} {incr i} {
    	    for {set j 0} {$j < $rows} {incr j} {
                .b($i,$j) configure -background blue
                if {$mine($i,$j)} {
                    .b($i,$j) configure -image flagImage -width $wImage -text F -foreground red 
                }
            }
        }  
    }
    if {$show==""} {
        set retval 0
        for {set i [expr $x-1]} {$i <= $x+1} {incr i} {
    	    for {set j [expr $y-1]} {$j <= $y+1} {incr j} {
                if {$i >= 0 && $i < $columns && $j >= 0 && $j < $rows && \
                    [info exists mine($i,$j)]} {reveal $i $j}
            }
        }
    }
}

# Called when a minefield is right-clicked to flag or unflag it,
# without revealing what is underneath
proc flag {x y} {
    global .b gameOver wImage wText minesLeft
    if {$gameOver} return
    set relief [lindex [.b($x,$y) configure -relief] 4]
    if {$relief != "raised"} return
    set text [lindex [.b($x,$y) configure -text] 4]
    if {$text == "F"} {
        .b($x,$y) configure -text "?" -image "" -width $wText -foreground blue
        incr minesLeft
    } elseif {$text == "?"} {
        .b($x,$y) configure -text "" -image "" -width $wText -foreground black 
    } else {
        .b($x,$y) configure -text "F" -image flagImage  -width $wImage -foreground red
        incr minesLeft -1
    }
}

# Repopulate the board with mines
proc newGame {} {
    global gameOver .b mine rows columns mines minesLeft revealed wText timer
    set gameOver 0
    set revealed 0
    set timer 0

    # Prepare the board
    for {set i 0} {$i < $columns} {incr i} {
    	for {set j 0} {$j < $rows} {incr j} {
            set mine($i,$j) 0
            .b($i,$j) configure -relief raised -text "" -image "" -width $wText -background gray -fg black
        }
    }

    # Place the mines
    for {set m 0} {$m < $mines} {} {
        set i [expr int($columns * rand())]
        set j [expr int($rows * rand())]
        if {!$mine($i,$j)} {
            set mine($i,$j) 1
            incr m
        }
    }
    set minesLeft $mines 
}

proc showAbout {} {
    toplevel .aboutBox
    wm title .aboutBox "About Minesweeper"
    message .aboutBox.msg -width 3i -text "George's Minesweeper Game,\
    \ndedicated to Alexandra!\n:-)"
    pack .aboutBox.msg
}

# Start play with a new board of a different size
proc newBoard {newColumns newRows newMines} {
    global rows columns mines .b .board wText wImage grinImage minesLeft

    # get rid of the old board first
    destroy .board
    destroy .b
    for {set i 0} {$i < $columns} {incr i} {
	    for {set j 0} {$j < $rows} {incr j} {
            destroy .b($i,$j)
        }
    }

    # build the new board
    set rows $newRows
    set columns $newColumns
    set mines $newMines

    frame .board
    pack .board -expand 1 -fill both

    frame .board.r0
    pack .board.r0 -expand 1 -fill both
    button .board.r0.grinButton -width $wImage -image grinImage -command newGame
    label .board.r0.minesLabel -textvariable minesLeft -foreground red -background black
    label .board.r0.timeLabel -textvariable timer -foreground red -background black
    pack  .board.r0.minesLabel .board.r0.grinButton .board.r0.timeLabel -side left -expand 1


    for {set i 0} {$i < $columns} {incr i} {
        frame .board.row($i)
        pack .board.row($i) -expand 1 -fill both
	    for {set j 0} {$j < $rows} {incr j} {
            button .b($i,$j) -width $wText -command "reveal $i $j"
            pack .b($i,$j) -in .board.row($i) -side left  -expand 1 -fill both
            bind .b($i,$j) <Button-3> "flag $i $j"
        }
    }

    # populate with mines
    newGame
}

# increments timer every second until game over
proc tick {} {
    global timer gameOver
    if { !$gameOver } {
        incr timer
    }
    after 1000 tick
}

image create photo flagImage -file flag.gif
image create photo mineImage -file mine.gif
image create photo grinImage -file grin.gif

frame .menubar
frame .board

menubutton .menubar.game -text Game -underline 0 -menu .menubar.game.menu
menu .menubar.game.menu -tearoff 0
.menubar.game.menu add command -label New -underline 0 -accelerator F2 -command newGame
bind all <F2> newGame
.menubar.game.menu add separator
.menubar.game.menu add radiobutton -label Beginner -underline 0 -command "newBoard 8 8 10" \
    -variable boardSize -value Beginner
.menubar.game.menu add radiobutton -label Intermediate -underline 0 -command "newBoard 16 16 40"\
    -variable boardSize -value Intermediate
.menubar.game.menu add radiobutton -label Expert -underline 0 -command "newBoard 16 30 99"\
    -variable boardSize -value Expert
# .menubar.game.menu add radiobutton -label Custom -underline 0 -command "newBoard 8 8 10"\
#    -variable boardSize -value Custom
.menubar.game.menu add separator
.menubar.game.menu add command -label Exit -underline 1 -command exit

menubutton .menubar.help -text Help -underline 0 -menu .menubar.help.menu
menu .menubar.help.menu -tearoff 0
.menubar.help.menu add command -label "About Minesweeper" -underline 0 -command showAbout

pack .menubar -expand 1 -fill x
pack .menubar.game .menubar.help -side left
tk_menuBar .menubar .menubar.game .menubar.help
focus .menubar

# start with "beginner" board
set boardSize Beginner
# start timing
after 1000 tick
newBoard 8 8 10


